/*
 * ALSA example app.
 * Based on tutorial at http://www.suse.de/~mana/alsa090_howto.html.
 */

#include <alsa/asoundlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <fftw3.h>
#include <signal.h>

#include "main.h"
#include "fourier.h"

/*********************
 * Global variables: *
 *********************/

// Handle for the PCM device:
snd_pcm_t *pcm_handle;

// Capture stream:
snd_pcm_stream_t stream = SND_PCM_STREAM_CAPTURE;

// Hardware parameters:
snd_pcm_hw_params_t *hwparams;

// Name of the PCM device, like plughw:0,0. The first number is the number of
// the soundcard, the second number is the number of the device.
char *pcm_name;

// ALSA configuration:
int direction = SND_PCM_ACCESS_RW_NONINTERLEAVED; // Access mode
int format = SND_PCM_FORMAT_S16_LE; // Format (16-bit unsigned little-endian)
int bytespersample = 2; // Depends on format (must match!)
int endianness = 1; // Depends on format (must match!)
int channels = 1; // Number of channels
int rate = 44100; // Sample rate
int periods = 2;  // Number of periods in the buffer
// (hardware interrupts occur at the edges of periods)
int periodsize = 8192; // The size of each period (in bytes)
                       // Note that number of periods + period
                       // size implicitly defines buffer size.

// The number of bytes to use for results (results are in Hz within human
// hearing range):
int peakwidth = 2;

// The number of peaks to package into a single message:
int batchsize = 3;

// The socket object that connects to the navi server:
int sock;

// Socket configuration:
char* host = "127.0.0.1";
int port = 51422;

// The disconnect signal:
const char* disconnect = "D";

// Help text:
const char* usage = \
"Usage: heylisten CARD\n"\
"  Takes PCM input from the given sound card (ALSA cardname, such as\n"\
"  'default' or 'plughw:0,0' required) and recognizes monophonic notes,\n"\
"  sending them to the navi server for recognition and command execution.\n";

/**************
 * Functions: *
 **************/

/*
 * The main function is where control starts, it also receives command-line
 * arguments. First, it sets up the PCM capture stream with a call to setupPCM,
 * then it starts the capture loop by calling processStream.
 */
int main (int argc, char** argv)
{
  int result;

  if (argc < 2) {
    fprintf(stderr, usage);
    return -1;
  }

  // Initialize pcm_name:
  pcm_name = strdup(argv[1]);

  // Allocate the snd_pcm_hw_params_t structure on the stack:
  snd_pcm_hw_params_alloca(&hwparams);

  // Set up the sound input stream:
  if ((result = setupPCM(&pcm_handle, stream, hwparams, pcm_name, direction,
                         format, channels, periods, &rate, &periodsize)) != 0) {
    fprintf(stderr, "Error during PCM setup. Aborting.\n");
    return result;
  }

  // Set up the FFT code:
  setupFourier(periodsize, channels, bytespersample, endianness, rate);

  // Set up the results socket:
  if ((result = setupSocket(host, port)) != 0){
    fprintf(stderr, "Error during socket setup. Aborting.\n");
    return result;
  }

  // Set up the signal handler for cleaning things up:
  signal(SIGINT, cleanup);

  // Enter the main sound-processing loop:
  processStream(pcm_handle, bytespersample, channels, rate,
                periods, periodsize);

  // Send the disconnect signal to the navi server:
  send(sock, disconnect, 1, 0);

  // Clean up the socket:
  shutdownSocket();

  // Clean up the Fourier stuff:
  teardownFourier();

  // Clean up the PCM:
  snd_pcm_drop(pcm_handle);

  return 0;
}

/*
 * This function sets up an ALSA PCM input or output stream according to the
 * following parameters. It sets up the pcm_handle passed to it, and returns
 * 0 on success, or -1 on failure. It may modify the rate or periodsize
 * passed to it to conform to underlying sound card restrictions.
 */
int setupPCM(snd_pcm_t **pcm_handle,
             snd_pcm_stream_t stream,
             snd_pcm_hw_params_t *hwparams,
             char *pcm_name,
             int direction,
             int format,
             int channels,
             int periods,
             int *rate,
             int *periodsize)
{
  // These variables hold exact values returned by _near methods.
  unsigned int exact_rate;
  snd_pcm_uframes_t exact_buffer_size, expected_buffer_size;

  // Open PCM. The last parameter of this function is the mode. If this is set
  // to 0, the standard mode is used. Possible other values are
  // SND_PCM_NONBLOCK and SND_PCM_ASYNC. If SND_PCM_NONBLOCK is used, read /
  // write access to the PCM device will return immediately. If SND_PCM_ASYNC
  // is specified, SIGIO will be emitted whenever a period has been completely
  // processed by the soundcard.
  if (snd_pcm_open(pcm_handle, pcm_name, stream, 0) < 0) {
    fprintf(stderr, "Error opening PCM device %s\n", pcm_name);
    return -1;
  }

  // Initialize hwparams with full configuration space:
  if (snd_pcm_hw_params_any(*pcm_handle, hwparams) < 0) {
    fprintf(stderr, "Can not configure this PCM device.\n");
    return -1;
  }

  // Set access type. This can be either SND_PCM_ACCESS_RW_INTERLEAVED or
  // SND_PCM_ACCESS_RW_NONINTERLEAVED. There are also access types for MMAPed
  // access, but this is beyond the scope of this introduction.
  if (snd_pcm_hw_params_set_access(*pcm_handle, hwparams, direction) < 0) {
    fprintf(stderr, "Error setting access.\n");
    return -1;
  }

  // Set sample format:
  if (snd_pcm_hw_params_set_format(*pcm_handle, hwparams, format) < 0) {
    fprintf(stderr, "Error setting format.\n");
    return -1;
  }

  // Set the sample rate. If the exact rate is not supported by the hardware,
  // use nearest possible rate.
  exact_rate = *rate;
  if (snd_pcm_hw_params_set_rate_near(*pcm_handle, hwparams,
                                      &exact_rate, 0) < 0) {
    fprintf(stderr, "Error setting rate.\n");
    return -1;
  }
  if (*rate != exact_rate) {
    fprintf(stderr, "The rate %d Hz is not supported by your hardware.\n"\
                    "==> Using %d Hz instead.\n", *rate, exact_rate);
    *rate = exact_rate;
  }

  // Set number of channels:
  if (snd_pcm_hw_params_set_channels(*pcm_handle, hwparams, channels) < 0) {
    fprintf(stderr, "Error setting channels.\n");
    return -1;
  }

  // Set number of periods:
  if (snd_pcm_hw_params_set_periods(*pcm_handle, hwparams, periods, 0) < 0) {
    fprintf(stderr, "Error setting periods.\n");
    return -1;
  }

  // Set buffer size (in frames). The resulting latency is given by
  // latency = periodsize * periods / (rate * bytes_per_frame). The buffer size
  // in frames is calculated by multiplying the period size (in bytes) by the
  // number of periods to get the size of the buffer and then dividing by the
  // number of bytes per frame (which is 2 bytes per channel for 16-bit audio).
  expected_buffer_size = (snd_pcm_uframes_t) (*periodsize * periods) /
                                                 (bytespersample * channels);
  exact_buffer_size = expected_buffer_size;
  if (snd_pcm_hw_params_set_buffer_size_near(*pcm_handle, hwparams,
                                             &exact_buffer_size) < 0) {
    fprintf(stderr, "Error setting buffersize.\n");
    return -1;
  }
  if (exact_buffer_size != expected_buffer_size) {
    fprintf(stderr, "The buffer size %d frames is not supported by your"\
                    "hardware.\n ==> Using %d frames instead.\n",
                    (int) expected_buffer_size, (int) exact_buffer_size);
    *periodsize = ((int) exact_buffer_size) * (bytespersample * channels) /
                  periods;
  }

  // Apply HW parameter settings to PCM device and prepare device:
  if (snd_pcm_hw_params(*pcm_handle, hwparams) < 0) {
    fprintf(stderr, "Error setting HW params.\n");
    return -1;
  }

  return 0;
}

/*
 * Sets up the socket (stored as a handle in the 'sock' variable) that
 * connects to the navi server at the given host and port. Port should
 * usually be 51422, and host should usually be localhost (127.0.0.1). This
 * function returns 0 on success and a negative number on failure.
 */
int setupSocket(char* host, int port)
{
  struct sockaddr_in server_address;
  int hostaddr = inet_addr(host);
  int result;

  // Create the socket:
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "Could not create socket.\n");
    return -1;
  }

  // Set the fields of the server_address:
  server_address.sin_family = AF_INET;
  server_address.sin_addr.s_addr = hostaddr;
  server_address.sin_port = htons(port);

  // Open the connection:
  if ((result = connect(sock, (struct sockaddr*) &server_address,
                        sizeof(server_address))) < 0) {
    fprintf(stderr, "Could not connect to server.\n");
    return result;
  }

  return 0;
}

/*
 * Shuts down the socket, closing the connection to the navi server.
 */
void shutdownSocket()
{
  if (close(sock) < 0) {
    fprintf(stderr, "Error closing socket.\n");
  }
}

/*
 * This function takes all of the specifics of the stream and enters an
 * infinite loop, grabbing periods one at a time from the sound card and
 * calling processPeriod on each. Make sure that setupFourier is called
 * before this function.
 */
void processStream(snd_pcm_t *pcm_handle,
                   int bytespersample,
                   int channels,
                   int rate,
                   int periods,
                   int periodsize)
{
  unsigned char *data; // Storage for data from the sound card
  int pcmreturn; // Result of the read (number of frames actually read)
  int frames; // The number of frames to read
  int result; // The result of the period processing
  int i, j; // Loop variables
  unsigned char *msg; // The message to send to the navi server.
  int msgindex = 1; // The partial index into the message.
  int *peaks; // Result from the Fourier processing.

  // Allocate memory:
  data = (unsigned char *) malloc(periodsize);
  msg = (unsigned char *) malloc(peakwidth * batchsize + 1);

  msg[0] = 'P'; // Message header (stands for "Peaks").

  // Calculate number of frames to request:
  frames = periodsize / (channels * bytespersample);

  // Enter the processing loop:
  while (1) {
    // Request (and receive) frames from the sound card:
    while ((pcmreturn = snd_pcm_readn(pcm_handle, (void**)&data, frames)) < 0) {
      snd_pcm_prepare(pcm_handle);
      fprintf(stderr, "<<<<<<<<<<<<<<< Buffer Overrun >>>>>>>>>>>>>>>\n");
    }
    if (pcmreturn != frames) {
      fprintf(stderr, "<<<<<<<<<<<<<<<< Input Underrun >>>>>>>>>>>>>>>\n");
    }

    // Process the sample:
    peaks = processPeriod(data);

    // DEBUG:
    // printf("P%d\n", peaks[0]);

    // DEBUG:
    /*
    for (i = 0; i < peaks[0]; ++i) {
      printf("%d ", peaks[i + 1]);
    }
    printf("\n");
    */

    // Loop through the returned peaks, bundling them into message batches and
    // sending them off:
    for (i = 0; i < peaks[0]; ++i) {
      
      // Put bytes into the message:
      for (j = 0; j < peakwidth; ++j) {
        // Big-endian unsigned message bytes:
        msg[msgindex + j] = (peaks[i + 1] >> (8 * (peakwidth - j - 1))) & 255;
      }
      // Keep track of position within the message:
      msgindex += peakwidth;

      // When the message is full, send it off and reset the index:
      if (msgindex > peakwidth * batchsize) {
        // Send the results batch to the navi server:
        if ((result = send(sock, msg, peakwidth * batchsize + 1, 0)) <
            peakwidth * batchsize + 1) {
          fprintf(stderr, "Warning: bad send with %d/%d bytes.\n", result,
                  peakwidth * batchsize + 1);
        }
        msgindex = 1;
      }

    }

    // Free the peaks data (malloced in processPeriod):
    free(peaks);
  }
}

void cleanup(int sig)
{
  // Send the disconnect signal to the navi server:
  send(sock, disconnect, 1, 0);

  // Clean up the socket:
  shutdownSocket();

  // Clean up the Fourier stuff:
  teardownFourier();

  // Clean up the PCM:
  snd_pcm_drop(pcm_handle);

  exit(EXIT_SUCCESS);
}
